(***************** Using the Lazy module ******************)
(* Here we provide an alternate implementation of streams using
   OCaml's lazy module. We recommend that you explore the
   documentation at
   http://caml.inria.fr/pub/docs/manual-ocaml/libref/Lazy.html

   In this portion, you will be reimplementing various functions that
   were defined in class and several more ...
*)

(* This file uses OCaml Num library for arbitrary precision arithmetic.
 *
 * To load this file in the OCaml toplevel, you need to start Ocaml 
 * with nums.cma (the Unix library) as an argument: 
 *
 *   % ocaml nums.cma
 *            Objective Caml version 3.11.0
 *   # 
 *
 * To use ocamlbuild, see the associated Makefile.
 *)

open Lazy;;

type 'a stream = Cons of 'a * 'a stream Lazy.t;;

let rec ones = Cons(1, lazy(ones));;

(*>* Problem 2.1.a *>*)
(* Implement the head function *)

let head (s:'a stream) : 'a =
  raise (Failure "unimplemented")
;;

(*>* Problem 2.1.b *>*)
(* Implement map *)

let rec map (f:'a -> 'b) (s:'a stream) : 'b stream =
  raise (Failure "unimplemented")
;;

(*>* Problem 2.1.c *>*)
(* Define nats *)

let rec nats = raise (Failure "unimplemented") ;;

(*>* Problem 2.1.d *>*)
(* Write a function nth, which returns the nth element of a
   stream. NOTE: the function nth should be zero-indexed. In other
   words, "nth 0 s" should be equivalent to "head s". *)

let rec nth (n:int) (s:'a stream) : 'a =
  raise (Failure "unimplemented")
;;

(*>* Problem 2.1.e *>*)
(* Now suppose we have two int streams s1 and s2 sorted in ascending
   order. We wish to merge these into a single stream s such that s is
   sorted in ascending order and has no duplicates. Implement this
   function. NOTE: This is not a simple merge function. You must
   REMOVE DUPLICATES.  *)

let merge (s1:int stream) (s2:int stream) : int stream =
  raise (Failure "unimplemented")
;;

(*>* Problem 2.1.f *>*)
(* What problems can we run into with this conception of "merge"? What
   if we were to run "merge ones ones"? Answer within the comment. *)

(* 
   Answer:
*)

(*>* Problem 2.1.g *>*)
(* Write a function "scale", which takes an integer "n" and an int
   stream "s", and multiplies each element of "s" by "n". *)

let scale n = raise (Failure "unimplemented") ;;

(*>* Problem 2.1.h *>*)
(* Suppose we wish to create a stream of the positive integers "n" in
   increasing order such that any prime factor of "n" is either 3 or
   5. The first few numbers are 1, 3, 5, 9, 15, 25, 27, ... hopefully
   you get the idea. One way to do this is to run "filter" over
   "nats". But we can do better than that. *)

(* Let "selectivestream" denote the desired stream. Observe that "selectivestream" satisfies the
   following properties

   1. The elements of "scale S 3" are elements of "selectivestream"
   2. The elements of "scale S 5" are elements of "selectivestream"
   3. "selectivestream" is the minimal stream (sorted in increasing order) which
   contains "1" and satisfies the above properties

   Think about why properties 1-3 uniquely characterize "selectivestream".
*)

(* Use the above discussion and functions you've already written to
   give a simple definition for "selectivestream". This can be done quite
   elegantly. *)

let rec selectivestream = raise (Failure "unimplemented") ;;

(*>* Problem 2.2 *>*)

(* Define a type for an infinite spreadsheet full of cells with type 'a. 
 *
 * You are free to use any representation of infinite spreadsheets you choose.
 *
 * Each cell in the spreadsheet will have coordinates (i,j).  You can think
 * of a cell with coordinates (i,j) as the cell inhabiting the ith row and
 * jth column of the spreadsheet.  You can assume that (i,j) are both
 * non-negative.  Indices in the spreadsheet start at 0.

 * Coordinates will be represented using OCaml's arbitrary precision Num
 * library.  See here:

   http://caml.inria.fr/pub/docs/manual-ocaml-4.00/libref/Num.html

   Note that you can use +/ -/ */ // for addition, subtraction, 
   multiplication and division on the type num.
 *)
(* Such a spreadsheet will need to support the following operations *)

open Big_int;;

type 'a spread_sheet = big_int;; (* change me! *)

(* we will pretend that ints are infinite precision *)
(* you can assume all coordinates given are non-negative *)
type coordinates = big_int * big_int ;;

(* a spreadsheet containing all zeros *)
let zeros : big_int spread_sheet = big_int_of_int 0;; (* change me *)

(* return the element at the (i,j) coordinate in ss *)
let get ((i,j):coordinates) (ss:'a spread_sheet) : 'a = 
  failwith "unimplemented"
;;

(* create a new spreadsheet where the (i,j) element of the spreadsheet
 * contains f i j xij  when xij was the (i,j) element of the input spreadsheet
 *)
let map_all 
    (f:big_int -> big_int -> 'a -> 'b) 
    (ss:'a spread_sheet) 
    : 'b spread_sheet = 
  failwith "unimplemented"
;;

(* create an infinite multiplcation table in which every cell contains the
 * product of its indices *)
let multiplication_table = big_int_of_int 0;; (* change me *)

(* produce a spreadsheet in which cell (i,j) contains the ith element
 * of is and the jth element of js *)
let cross_product (is:'a stream) (js:'b stream) : ('a * 'b) spread_sheet =
  failwith "unimplemented"
;;

(* KARMA (Optional): Cryptographic keys are often generated from products
 * of large prime big_intbers.  Use cross_product and map_all to help you
 * compute a spreadsheet that for all pairs of primes p and q contains
 * a square in the spreadsheet that contains p*q.
 * Recall that to generate all primes, you can use an algorithm 
 * called the Sieve of Eratosthenes.
 *
 * see here: 
 * http://en.wikipedia.org/wiki/Sieve_of_Eratosthenes
 * http://www.cs.princeton.edu/~dpw/courses/cos326-12/lec/code/streams.tgz
 *)

let product_of_primes : big_int spread_sheet = big_int_of_int 0;; (* change *)



